/* lcdsubs.c
 * Copyright (C) 2013-2014 Jim Jackson  <jj@franjam.org.uk>
 *
 * LCD subs that depend on an understanding of the LCD hardware
 * connection. 
 *
 * This version is for an HD44780 LCD connected to an i2c MCP23017 chip
 * using Jim Jackson's MCP23017 breakout board.
 *
 * The connections between the MCP23017 and the LCD are:
 *   data portb connected to the 8 LCD data pins
 *   data porta D7 to LCD E   Enable[1]     
 *   data porta D6 to LCD R/W Read[1]/Write[0]
 *   data porta D5 to LCD RS  Data[1]/Command[0]
 *   data porta D4 to LCD Backlight  On[1]/Off[0]
 *   data porta D3 to Sounder On[1]/Off[0]
 *
 */
/*
 *  Date       Vers   Comment
 *  -----      -----  ---------------
 *  2014-01-02  1     
 *
 *  This contains the functions:
 *    int lcd_init((char *dev, int add, int ifw, int l)
 *                          initialise lcd - not all args may be used by 
 *                          all drivers.
 *    int lcd_put(int c)    sends lower 8 bits of c to lcd as a data byte
 *    int lcd_puts(char *s,N) send string in s, or N bytes to lcd as data
 *    int lcd_cmd(int c)    sends lower 8 bits of c to lcd as a cmd byte
 *    int lcd_bel(int period, int type)  do the necessary for the BEL code
 *                          in this driver it flashes the LCD backlight
 *                          off for period millisecs.
 *    int lcd_backlight(int v)  set backlight - v is driver dependant
 *                          in this version v=0 is off, else backlight on
 *
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program - see the file COPYING; if not, write to 
 *  the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 *  MA 02139, USA.
 */

/* INCLUDES .................................................... */

#include <stdlib.h>  
#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <ctype.h>
#include "lcddriver.h"
#include "timersubs.h"

#define IODIRA    (0x00)
#define IODIRB    (0x01)
#define CNTL_PORT (0x12)
#define DATA_PORT (0x13)

int lcd_ifwidth=8;
int lcd_lines=0;
int lcd_fd=-1;

unsigned int gpcounter=0;   /* a counter for setting a pause */

/* CNTL_PORT bits and their uses...
 */
#define EN_BIT    (0x80)    /* Enable pulse 1=on to read write data */
#define RW_BIT    (0x40)    /* Read/Write  1=read  0=write */
#define RS_BIT    (0x20)    /* RS bit 1=data  0=command */
#define BL_BIT    (0x10)    /* Backlight bit 1=on */
#define SND_BIT   (0x08)    /* Sounder control 1=on */

int lcd_control=BL_BIT;  /* These bits are OR'ed into any CNTL_PORT writes
                            They control ancillary hardware functions */

int LCD_LOCK=0;    /* Global Lock bit. When set then one of the 
                      functions here is dealing with the LCD and does not
                      want to be interrupted. Here to prevent any timer
                      function from messing things up. */

/* lcd_init(dev,add,ifw,l)  open i2c connected LCD on i2c device dev and,
 *    if add > 0 , set the i2c device address to add.
 *    this only uses 8 bit i/f. l is used for singleline or multiline setup
 *    return 0 if all ok else return errno
 */

int lcd_init(char *dev,int add, int ifw, int l) {
  int fd,st;

  if (lcd_fd==-1) {   /* if i2c not opened then open it it else just 
                         re-initialise it */
    fd = i2c_init(dev,add);
    if (fd == -1) { return(errno); }

    lcd_lines=(l>1)?2:1;

  /* set both data ports as outputs by setting data io directions to zero 
   * and set all outputs to zero and initialise gpcounter.
   * if any error then return it and close fd
   */

    if ( (st=i2c_setreg(fd,IODIRA,0)) ||
         (st=i2c_setreg(fd,IODIRB,0)) ||
         (st=i2c_setreg(fd,DATA_PORT,0)) ||
         (st=i2c_setreg(fd,CNTL_PORT,0)) ||
         (st=counter_setup(&gpcounter,0,NULL)) ) { 
      close(fd);   
      return(st); 
    }
    lcd_fd=fd;
  }

  /* initialise the display ... */
  lcd_cmd(0x38);
  lcd_wait(10); 
  lcd_cmd(0x38);  /* again for good luck */
  lcd_wait(10); 
  lcd_cmd(0x0C+DEFAULT_CURSORMODE);  /* set display on, default cursor mode */
  lcd_cmd(0x06);  /* increment mode, entire shift off */
  lcd_cmd(0x01);  /* clear display */
  lcd_backlight(DEFAULT_BACKLIGHT); /* set default backlight state */

  return(0);
}

/* lcd_cmd(c)  send command "c" to lcd
 * return 0 if ok else errno
 *
 * set the LOCK, then set the cmd on the data port, then set the enable bit,
 * then clear it - remembering to add in the other control bits as set.
 */

int lcd_cmd(int c) {
  int st;

  LCD_LOCK++;
  (st=i2c_setreg(lcd_fd,DATA_PORT,c)) || 
    (st=i2c_setreg(lcd_fd,CNTL_PORT,lcd_control | EN_BIT)) ||
    (st=i2c_setreg(lcd_fd,CNTL_PORT,lcd_control));
  LCD_LOCK--;
  if ( c==1 || c==2 ) { lcd_wait(5); }
  return(st);

}

/* lcd_put(c)  send data "c" to lcd
 * return 0 if ok else errno
 *
 * set the LOCK, then set the char on the data port, then set the enable bit
 * and the RS bit, then clear the enable bit - remembering to add in the 
 * other control bits as set.
 */

int lcd_put(int c) {
  int st;

  LCD_LOCK++;
  (st=i2c_setreg(lcd_fd,DATA_PORT,c)) ||   
    (st=i2c_setreg(lcd_fd,CNTL_PORT,lcd_control | EN_BIT | RS_BIT)) ||
    (st=i2c_setreg(lcd_fd,CNTL_PORT,lcd_control | RS_BIT));
  LCD_LOCK--;
  return(st);

}

/* lcd_puts(s,n)  send string in s to lcd as data - Max. N bytes sent
 */

int lcd_puts(s,n)
unsigned char *s;
int n;
{
  int i;

  for (i=0; i<n && s[i] ; i++) { lcd_put(s[i]); }
}

/* lcd_bel(period)  do this drivers implementation of a BEL
 *   VISUAL - toggle backlight.
 *   AUDIO  - turn on for period.
 */

int lcd_bel(int period,int type)
{
  int st=0;

  if ( period>0 ) {
    // visual - toggle backlight ...
    lcd_control ^= (type & VISUAL_BELL) ? BL_BIT : 0 ; 
    // sounder - on then off ...
    lcd_control |= (type & AUDIO_BELL) ? SND_BIT : 0 ;   
    st=i2c_setreg(lcd_fd,CNTL_PORT,lcd_control);
    gpcounter=period;
    waitcounter(gpcounter);
    lcd_control ^= (type & VISUAL_BELL) ? BL_BIT : 0 ;  // toggle back
    lcd_control &= (type & AUDIO_BELL) ? (~SND_BIT) : 0xFF ;  // sounder off
    st=i2c_setreg(lcd_fd,CNTL_PORT,lcd_control);
  }
  return(st);
}

/* lcd_backlight(value)  action the backlight depending on value
 *  In this version simply set backlight on, value!=0, or off, value==0.
 */

int lcd_backlight(int value) {

  if (value) {
    lcd_control|=BL_BIT;
  } else {
    lcd_control&=(~BL_BIT);
  }
  return(i2c_setreg(lcd_fd,CNTL_PORT,lcd_control));
}

void lcd_wait(unsigned int p) {
  if (p) {
    gpcounter=p;
    while (gpcounter) { waitalarm(); }
  }
}
