#!/bin/bash
#
##
## Usage: _PROG_ [options] file(s)
##
## Used to test the ANSI LCD driver. Interprets commands read from file(s) 
## or STDIN and sends relevant control codes to _DEVICE_.
##
## Commands:
##  reset        send ESC 'c' to reset the device
##  bel          send 0x07 (\007) 
##  bs           send 0x08 (\010)  backspace cursor
##  lf           send 0x0A (\012)  linefeed
##  ff           send 0x0C (\014)  formfeed (clear screen)
##  cr           send 0x0D (\015)  carriage return (cursor to start of line
##  so           send 0x0E (\016)  next byte is an HD44780 Control byte
##  esc          send 0x1B (\033)  send escape char
##  csi          send 0x9B (\233)  send csi (same as ESC [ )
##  lw           send 0x1B [ 7 h   Enable Line Wrap
##  nolw         send 0x1B [ 7 l   Disable Line Wrap
##  fsg0         send 0x1B (       Font Set G0
##  fsg1         send 0x1B )       Font Set G1
##  scrdn        send 0x1B D       Scroll Down
##  scrup        send 0x1B M       Scroll UP
##  cudN         send 0x1B [ N A   Cursor down N lines
##  cuuN         send 0x1B [ N B   Cursor up   N lines
##  curN         send 0x1B [ N C   Cursor right N lines
##  culN         send 0x1B [ N D   Cursor left N lines
##  eeos         send 0x1B [ 0 J   Erase to End Of Screen
##  esos         send 0x1B [ 1 J   Erase to Start Of Screen
##  eaos         send 0x1B [ 2 J   Erase All Of Screen
##  eeol         send 0x1B [ 0 K   Erase to End Of Line
##  esol         send 0x1B [ 1 K   Erase to Start Of Line
##  eaol         send 0x1B [ 2 K   Erase All Of Line
##  home         send 0x1B [ H     Cursor to top left of screen
##  @L;C         send 0x1B [ L ; C H
##                                 position cursor at line L and col C
##  scu          send 0x1B [ s     save cursor position
##  rcu          send 0x1B [ r     restore cursor position
##  blon         send 0x1B [ 1 V   backlight on
##  bloff        send 0x1B [ V     backlight off
##  blN          send 0x1B [ N V   backlight value N
##  emlr         send 0x1B [ 4 E   Entry Mode Left Reverse
##  emrr         send 0x1B [ 5 E   Entry Mode Right Reverse
##  emln         send 0x1B [ 4 E   Entry Mode Left Normal
##  emrn         send 0x1B [ 4 E   Entry Mode Right Normal
##  cursoff      send 0x1B [ 0 X   Cursor Off
##  curson       send 0x1B [ 2 X   Cursor Underscore On
##  cursfl       send 0x1B [ 3 X   Cursor Flash On
##  dispoff      send 0x1B [ 8 X   Display Off
##  dispon       send 0x1B [ 12 X  Display On
##
## Options:
##  -d device  device to write to. Default is /dev/lcd
##  -h    get this help message
##  -v    verbose
##
#

prog="$0"
me=`basename "$0"`
user=`id -un`
verbose=false
noexe=""
dev=/dev/lcd
HISTFILE=$HOME/.lcdtest_history

#===================================================================

dohelp () {
  pfx="$1"
  if [ "$pfx" = "" ]; then pfx='##' ; fi
  grep "^$pfx" "$prog" | \
   sed -e "s/^$pfx//" -e "s/_PROG_/$me/" -e "s~_DEVICE_~$dev~" 1>&2
}

#===================================================================

while [ "$1" != "" ]; do
 case "$1" in
 -h|--help)
	dohelp
 	exit 1
	;;
 -v)
	verbose=true
	;;
 -n)
	noexe=echo
	;;
 -d)
	if [ "$2" != "" ]; then
	  shift; dev="$1"
        fi
	;;
 -*)
	echo 1>&2 "Option \"$1\" not recognised."
	dohelp ; exit 1
	;;
 *)
	break;
	;;
 esac
 shift
done

files=true
if [ "$1" = "" ]; then files=false; fi

[ "$PS1" ] && history -r

exec 1>"$dev"
while true; do 

  if $files ; then
      fn="$1" ; shift
      exec < "$fn"
  fi

  while read -e -a cmds ; do

    [ "PS1" ] && history -s "${cmds[@]}"

    ((max=${#cmds[@]}, maxi=max - 1))
    for (( i=0; i <= maxi; i++ )); do

      cmd=${cmds[i]}
      $verbose && echo 1>&2 "--- $cmd"
      case "$cmd" in
	  reset|RESET)   printf "\033c" ;;
	  bel|BEL)       printf "\007"  ;;
	  bs|BS)         printf "\010"  ;;
	  lf|LF)         printf "\012"  ;;
	  ff|FF)         printf "\014"  ;;
	  cr|CR)         printf "\015"  ;;
	  so|SO)         printf "\016"  ;;
	  esc|ESC)       printf "\033"  ;;
	  csi|CSI)       printf "\233"  ;;
	  lw)            printf "\033[7h" ;;
	  nolw)          printf "\033[7l" ;;
	  fsg0)          printf "\033(" ;;
	  fsg1)          printf "\033)" ;;
	  scrdn)         printf "\033D" ;;
	  scrup)         printf "\033M" ;;
	  cursoff)       printf "\033[0X" ;;
	  curson)        printf "\033[2X" ;;
	  cursfl)        printf "\033[3X" ;;
	  cud*)          printf "\033[${cmd:3}A" ;;
	  cuu*)          printf "\033[${cmd:3}B" ;;
	  cur*)          printf "\033[${cmd:3}C" ;;
	  cul*)          printf "\033[${cmd:3}D" ;;
	  eeos)          printf "\033[0J" ;;
	  esos)          printf "\033[1J" ;;
	  eaos)          printf "\033[2J" ;;
	  eeol)          printf "\033[0K" ;;
	  esol)          printf "\033[1K" ;;
	  eaol)          printf "\033[2K" ;;
	  home)          printf "\033[H"  ;;
	  @*)            printf "\033[${cmd:1}H" ;;
	  scu)           printf "\033[s"  ;;
	  rcu)           printf "\033[r"  ;;
	  blon)          printf "\033[1V" ;;
	  bloff)         printf "\033[V" ;;
	  bl*)           printf "\033[${cmd:2}V" ;;
	  emlr)          printf "\033[4E" ;;
	  emrr)          printf "\033[5E" ;;
	  emln)          printf "\033[6E" ;;
	  emrn)          printf "\033[7E" ;;
	  dispoff)       printf "\033[8X" ;;
	  dispon)        printf "\033[12X" ;;
	  \"*\"|\'*\'|.*)
	                 printf -- "${cmd:1}" ;;
	  *)
	      echo "\"$cmd\" unrecognised." 1>&2
	      ;;
      esac

    done

  done

  [ "PS1" ] && history -w

  if [ "$1" = "" ] ; then break ; fi

done

exit 0
