/* timersubs.c
 * Copyright (C) 2013-2014 Jim Jackson  <jj@franjam.org.uk>
 *
 * General Purpose Timer function    see also timersubs.h
 *
 *  Date       Vers   Comment
 *  -----      -----  ---------------
 *  2014-01-03  1     
 */

/* 
 * setitimer is used to set up a repetitive timer function that executes
 * every so many millisecs as specified in the call to init_timers()
 *
 * The SIGALARM handler, timer_handler(), loops through an array of
 * th_counter structures, decrementing the counter if it is set (non-zero).
 *
 * timersubs.h  contains prototype for the counter_setup and 
 *    counter_remove functions. It also contains MAX_COUNTERS, which 
 *    sets how many counter structures are contained in the array.
 *    
 * Each th_counter structure contains 3 parts:
 *   1 - an unsigned int * counter  to the counter variable 
 *        if this is non-zero, it is decremented by the number of millisecs
 *        between  time handler calls. If this is a null pointer then
 *        this array entry is unused.
 *   2 - an unsigned int repcount  if non-zero then this is used to
 *        set the above counter everytime it decrements to zero.
 *        This is used to create a repeating action
 *        If this is zero then the counter is not reset. It is assumed the
 *        counter is set when it is needed outside of this code.
 *   3 - a void (*function()) - pointer to a function to execute when
 *        the counter is decremented to zero. If NULL then no function 
 *        is called.
 *
 * Code using the timer facility calls counter_setup(), with the 3 
 * parameters needed to fill a th_counter structure.
 *
 * counter_remove can be used to remove a counter entry. It is called with
 * the address of the counter variable. 
 *
 * If code needs a counter for single time intervals then it could
 * do the following...
 *
 *   unsigned int downcounter=0;
 *
 *   if (counter_setup(&downcounter,0,NULL)) {
 *      **** error branch ****
 *   }
 *   ...
 *   * wait 100 millisecs *
 *   downcounter=100; 
 *   ...do stuff...
 *   while(downcounter) { waitalarm(); }
 * 
 * the waitalarm() call suspends the process untill the next timer call.
 * This uses a lot less cpu then just checking is downcounter is zero in 
 * a tight loop - do NOT do this:    while (downcounter) { }
 * 
 * A function to be called every 200millisecs can be actioned like this
 *
 *   repfunction() {
 *     ...some code...
 *   }
 *
 *   unsigned int repcounter;
 *
 *   if (counter_setup(&repcounter,200,&repfunction())) {
 *      **** error branch ****
 *   }
 *
 */
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program - see the file COPYING; if not, write to 
 *  the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 *  MA 02139, USA.
 */

/* INCLUDES .................................................... */

#include <stdlib.h>  
#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <signal.h>
#include <sys/time.h>
#include <errno.h>   /* Error number definitions */
#include <ctype.h>
#include "timersubs.h"

struct sigaction sa;
struct itimerval timer;

unsigned int millisecs=0;  /* holds the number of millisecs that timer_handler
                              decrements each non-zero counter each time it is 
			      called. If the counter is less than millisecs
			      the counter is set to zero. 
			      It is initialised by init_timers()  */

typedef struct { 
  unsigned int *cptr;   /* pointer to the counter variable */
  unsigned int msecs;   /* if non-zero, then reset counter to this value
                           when it decrements to zero */
  void (*function)();   /* function to call when counter decrements to zero
                           or NULL */
} th_counter;

th_counter th_counters[MAX_COUNTERS]; /* the counter array
					 initialised by init_timers()  */


/* The repetitive timer code
 */

void timer_handler(int sig) {
  unsigned int i,*cp;

  for ( i=0; i<MAX_COUNTERS; i++) {
    cp=th_counters[i].cptr;
    if (cp != NULL && *cp) {
      *cp=(*cp>millisecs)?*cp-millisecs:0;
      if ( *cp == 0) {
	*cp = th_counters[i].msecs;
	if ( th_counters[i].function != NULL ) {
	  (th_counters[i].function)();
	}
      }
    }
  }
}

/* init_timers(ms)  setup itimer to call timer_handler every ms millisecs
 *  on error -1 is returnd and errno is set
 *  if all ok then zero is returned
 */

int init_timers(unsigned int ms) {
  unsigned int secs,usecs,i;

  millisecs=ms;
  secs=ms/1000;                    /* calculate number of seconds */
  usecs=(ms%1000)*1000;            /* and number of microsecs     */

  for ( i=0; i<MAX_COUNTERS; i++) { /* set the counters to unset conditions */
    th_counters[i].cptr=NULL;
    th_counters[i].msecs=0;
    th_counters[i].function=NULL;
  }

  memset (&sa, 0, sizeof (sa));    /* setup the repetitive timer_handler */
  sa.sa_handler = &timer_handler;
  sigaction (SIGALRM, &sa, NULL);
  timer.it_value.tv_sec = secs;    /* setup requested initial period */
  timer.it_value.tv_usec = usecs; 
  timer.it_interval.tv_sec = secs; /* and as repeat period */
  timer.it_interval.tv_usec = usecs;
  return(setitimer(ITIMER_REAL, &timer, NULL));
}

/* waitalarm()
 *  sigsuspend until the next alarm signal
 */

void waitalarm()
{
  sigset_t ss; 

  sigemptyset(&ss);
/*  sigaddset (&ss, SIGALRM);  */
  sigsuspend(&ss);
}

/* counter_setup(cp,ms,func)  find a free slote in the th_counters array
 *     and populate with the given parameters and set *cp to ms.
 *     If ms > 0 this starts a repeating counter.
 *     if no free slot return ENOSPC, if all ok return 0
 */

int counter_setup(unsigned int *cp, unsigned int ms, void (*func)()) {
  int i;

  if (cp==NULL) { return(0); }
  for ( i=0 ; i<MAX_COUNTERS; i++ ) {
    if (th_counters[i].cptr == NULL) {
      th_counters[i].function=func;
      th_counters[i].msecs=ms;
      th_counters[i].cptr=cp;
      *cp=ms;
      return(0);
    }
  }
  return(ENOSPC);
}

/* counter_setup(cp,ms,func)  find a free slote in the th_counters array
 *     and populate with the given parameters.
 *     if no free slot return ENOSPC, if all ok return 0
 */

int counter_remove(unsigned int *cp) {
  int i;

  if (cp==NULL) { return(0); }
  for ( i=0 ; i<MAX_COUNTERS; i++ ) {
    if (th_counters[i].cptr == cp) {
      th_counters[i].cptr=NULL;
    }
  }
  return(0);
}

